﻿#region "SoapBox.Core License"
/// <header module="SoapBox.Core"> 
/// Copyright (C) 2009 SoapBox Automation Inc., All Rights Reserved.
/// Contact: SoapBox Automation Licencing (license@soapboxautomation.com)
/// 
/// This file is part of SoapBox Core.
/// 
/// Commercial Usage
/// Licensees holding valid SoapBox Automation Commercial licenses may use  
/// this file in accordance with the SoapBox Automation Commercial License
/// Agreement provided with the Software or, alternatively, in accordance 
/// with the terms contained in a written agreement between you and
/// SoapBox Automation Inc.
/// 
/// GNU Lesser General Public License Usage
/// SoapBox Core is free software: you can redistribute it and/or modify 
/// it under the terms of the GNU Lesser General Public License
/// as published by the Free Software Foundation, either version 3 of the
/// License, or (at your option) any later version.
/// 
/// SoapBox Core is distributed in the hope that it will be useful, 
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU Lesser General Public License for more details.
/// 
/// You should have received a copy of the GNU Lesser General Public License 
/// along with SoapBox Core. If not, see <http://www.gnu.org/licenses/>.
/// </header>
#endregion
        
using System;
using System.Text;
using System.Collections.Generic;
using System.Linq;
using NUnit.Framework; 
using System.ComponentModel;
using System.Collections.ObjectModel;
using System.Windows.Media.Imaging;
using System.Windows.Controls;
using System.Drawing;
using System.Windows.Media;

namespace SoapBox.Core
{
    [TestFixture]
    public class AbstractMenuItemTest
    {

        #region " Header "
        [Test]
        public void TestHeader()
        {
            //test default value
            ConcreteMenuItem target = new ConcreteMenuItem();
            Assert.AreEqual(string.Empty, target.Header);

            //test setting value and property changed event
            bool propertyChanged = false;
            target.PropertyChanged += delegate(object sender, PropertyChangedEventArgs e)
                {
                    if (e.PropertyName == "Header")
                    {
                        Assert.IsFalse(propertyChanged);
                        propertyChanged = true;
                    }
                    else
                    {
                        throw new Exception("Unexpected Property Changed.");
                    }
                };
            string test = Guid.NewGuid().ToString();
            target.SetHeader(test);
            Assert.AreEqual(test, target.Header);
            Assert.IsTrue(propertyChanged);
        }

        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestHeaderNull()
        {
            //should throw exception on null value
            ConcreteMenuItem target = new ConcreteMenuItem();
            target.SetHeader(null);
        }
        #endregion

        #region " Items "
        [Test]
        public void TestItems()
        {
            //test default value
            ConcreteMenuItem target = new ConcreteMenuItem();
            Assert.IsNull(target.Items);

            //test setting value and property changed event
            bool propertyChanged = false;
            target.PropertyChanged += delegate(object sender, PropertyChangedEventArgs e)
            {
                if (e.PropertyName == "Items")
                {
                    Assert.IsFalse(propertyChanged);
                    propertyChanged = true;
                }
                else
                {
                    throw new Exception("Unexpected Property Changed.");
                }
            };
            Collection<IMenuItem> c = new Collection<IMenuItem>();
            IMenuItem temp = new ConcreteMenuItem();
            c.Add(temp); // just do this so it doesn't become invisible (due to no submenuitems)
            target.SetItems(c);
            Assert.AreEqual(c, target.Items);
            Assert.IsTrue(propertyChanged);
            Assert.IsTrue(target.Visible);
        }

        [Test]
        public void TestItems_EmptySubmenu()
        {
            //test default value
            ConcreteMenuItem target = new ConcreteMenuItem();
            Assert.IsNull(target.Items);

            //test setting value and property changed event
            bool propertyChanged_Items = false;
            bool propertyChanged_Visible = false;
            target.PropertyChanged += delegate(object sender, PropertyChangedEventArgs e)
            {
                if (e.PropertyName == "Items")
                {
                    Assert.IsFalse(propertyChanged_Items);
                    propertyChanged_Items = true;
                }
                else if (e.PropertyName == "Visible")
                {
                    Assert.IsFalse(propertyChanged_Visible);
                    propertyChanged_Visible = true;
                }
                else
                {
                    throw new Exception("Unexpected Property Changed.");
                }
            };
            Collection<IMenuItem> c = new Collection<IMenuItem>();
            Assert.IsTrue(target.Visible);
            target.SetItems(c); // empty submenu so we should see the visibility change to false
            Assert.AreEqual(c, target.Items);
            Assert.IsTrue(propertyChanged_Items);
            Assert.IsTrue(propertyChanged_Visible);
            Assert.IsFalse(target.Visible);
        }

        #endregion

        #region " Icon "
        [Test, RequiresSTA]
        public void TestIcon()
        {
            //test default value
            ConcreteMenuItem target = new ConcreteMenuItem();
            Assert.IsNull(target.Icon);

            //test setting value and property changed event
            int propertyChanged = 0;
            bool EnableConditionChanged = false;
            target.PropertyChanged += delegate(object sender, PropertyChangedEventArgs e)
            {
                if (e.PropertyName == "Icon")
                {
                    propertyChanged += 1;
                }
                else if (e.PropertyName == "EnableCondition")
                {
                    Assert.IsFalse(EnableConditionChanged);
                    EnableConditionChanged = true;
                }
                else
                {
                    throw new Exception("Unexpected Property Changed: " + e.PropertyName);
                }
            };
            Bitmap b = new Bitmap(1,1);
            target.SetIconFromBitmap(b);
            Assert.IsNotNull(target.Icon);
            Assert.AreEqual(2, propertyChanged);
            Assert.IsTrue(EnableConditionChanged);

        }

        [Test, RequiresSTA]
        public void Test_disabled_icon()
        {
            ConcreteMenuItem target = new ConcreteMenuItem();
            Bitmap b = new Bitmap(1, 1);
            target.SetIconFromBitmap(b);

            var iconFull = target.Icon;
            target.SetEnableCondition(new ConcreteCondition(false));
            var iconGray = target.Icon;

            Assert.AreNotSame(iconFull, iconGray);

            Assert.Throws<ArgumentNullException>(() => target.SetIconFromBitmap(null));
        }
        #endregion

        #region " IsCheckable "
        [Test]
        public void TestIsCheckable()
        {
            //test default value
            ConcreteMenuItem target = new ConcreteMenuItem();
            Assert.IsFalse(target.IsCheckable);

            //test setting value and property changed event
            bool propertyChanged = false;
            target.PropertyChanged += delegate(object sender, PropertyChangedEventArgs e)
            {
                if (e.PropertyName == "IsCheckable")
                {
                    Assert.IsFalse(propertyChanged);
                    propertyChanged = true;
                }
                else
                {
                    throw new Exception("Unexpected Property Changed.");
                }
            };
            target.SetIsCheckable(true);
            Assert.IsTrue(target.IsCheckable);
            Assert.IsTrue(propertyChanged);
        }
        #endregion

        #region " IsChecked "
        [Test]
        public void TestIsChecked()
        {
            //test default value
            ConcreteMenuItem target = new ConcreteMenuItem();
            Assert.IsFalse(target.IsChecked);

            //test setting value and property changed event
            bool propertyChanged = false;
            target.PropertyChanged += delegate(object sender, PropertyChangedEventArgs e)
            {
                if (e.PropertyName == "IsChecked")
                {
                    Assert.IsFalse(propertyChanged);
                    propertyChanged = true;
                }
                else
                {
                    throw new Exception("Unexpected Property Changed.");
                }
            };

            bool checkChanged = false;
            target.IsCheckedChanged += delegate(object sender, EventArgs e)
            {
                Assert.IsFalse(checkChanged);
                checkChanged = true;
            };

            ((IMenuItem)target).IsChecked=true;
            Assert.IsTrue(target.IsChecked);
            Assert.IsTrue(propertyChanged);
            Assert.IsTrue(checkChanged);
        }
        #endregion

        #region " IsSeparator "
        [Test]
        public void TestIsSeparator()
        {
            //test default value
            ConcreteMenuItem target = new ConcreteMenuItem();
            Assert.IsFalse(target.IsSeparator);

            //test setting value and property changed event
            bool propertyChanged = false;
            target.PropertyChanged += delegate(object sender, PropertyChangedEventArgs e)
            {
                if (e.PropertyName == "IsSeparator")
                {
                    Assert.IsFalse(propertyChanged);
                    propertyChanged = true;
                }
                else
                {
                    throw new Exception("Unexpected Property Changed.");
                }
            };
            target.SetIsSeparator(true);
            Assert.IsTrue(target.IsSeparator);
            Assert.IsTrue(propertyChanged);
        }
        #endregion

        #region " Context "
        [Test]
        public void TestContext()
        {
            //test default value
            ConcreteMenuItem target = new ConcreteMenuItem();
            ConcreteMenuItem targetChild = new ConcreteMenuItem();
            List<IMenuItem> items = new List<IMenuItem>();
            items.Add(targetChild);
            target.SetItems(items);
            Assert.AreEqual(null, target.Context);
            Assert.AreEqual(null, target.Items.First().Context);

            //test setting value and property changed event
            bool propertyChanged = false;
            target.PropertyChanged += delegate(object sender, PropertyChangedEventArgs e)
            {
                if (e.PropertyName == "Context")
                {
                    Assert.IsFalse(propertyChanged);
                    propertyChanged = true;
                }
                else
                {
                    throw new Exception("Unexpected Property Changed.");
                }
            };
            object test = new object();
            target.Context = test;
            Assert.AreEqual(test, target.Context);
            Assert.AreEqual(test, target.Items.First().Context);
            Assert.IsTrue(propertyChanged);
        }

        #endregion

        #region " IsSubmenuOpen "
        [Test]
        public void TestIsSubmenuOpen()
        {
            //test default value
            ConcreteMenuItem target = new ConcreteMenuItem();
            Assert.IsFalse(target.IsSubmenuOpen);

            //test setting value and property changed event
            bool propertyChanged = false;
            target.PropertyChanged += delegate(object sender, PropertyChangedEventArgs e)
            {
                if (e.PropertyName == "IsSubmenuOpen")
                {
                    Assert.IsFalse(propertyChanged);
                    propertyChanged = true;
                }
                else
                {
                    throw new Exception("Unexpected Property Changed.");
                }
            };

            bool checkChanged = false;
            target.IsSubmenuOpenChanged += delegate(object sender, EventArgs e)
            {
                Assert.IsFalse(checkChanged);
                checkChanged = true;
            };

            ((IMenuItem)target).IsSubmenuOpen = true;
            Assert.IsTrue(target.IsSubmenuOpen);
            Assert.IsTrue(propertyChanged);
            Assert.IsTrue(checkChanged);
        }
        #endregion
    }
}
